import argparse
from multiprocessing import Pool, Process, Manager
from pixiu.pxtester import PXTester
from tabulate import tabulate
from ctypes import c_wchar_p
import json
import uuid
import hashlib
import traceback
import os
from datetime import datetime


class MainApp:
    def __init__(self, args):
        self.test_names = []
        self.data_file = args.datafile
        self.set_tag(args.tag)

    @staticmethod
    def str2bool(v):
        if isinstance(v, bool):
           return v
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')

    def load_data(self):
        try:
            with open(f"pxdata/{self.data_file}", "r") as f:
                s = f.read()
                return json.loads(s)
        except:
            traceback.print_exc()
        return None

    def save_data(self, data):
        try:
            with open(f"pxdata/{self.data_file}", "w") as f:
                f.write(json.dumps(data))
            return True
        except:
            traceback.print_exc()
        return False

    def load_tag_data(self, tag):
        if not tag:
            return None
        d = self.load_data()
        if d is None:
            d = {'version': '0.1.0', 'tags': {}}
        return d['tags'][tag]

    def save_tag_data(self, tag, data):
        if not tag:
            return
        d = self.load_data()
        if d is None:
            d = {'version': '0.1.0', 'tags': {}}
        d['tags'][tag] = data
        self.save_data(d)

    def set_tag(self, tag):
        if tag is None:
            self.tag = hashlib.md5(str(uuid.uuid4()).encode("utf-8")).hexdigest()
        else:
            self.tag = tag

    def output_report(self, reports, compare_reports=[]):
        report_str = f"\n-- Result (Tag: {self.tag}) --\n"
        idx = 1
        data = []
        headers = self.test_names

        for key in reports:
            item = reports[key][self.test_names[0]]
            precision = item.get('precision', 2)
            item_type = item.get('type', 'value')
            # item_aggr = item.get('aggr', None)
            report_str += f"{idx:02d}). {item['desc']}: "
            if len(compare_reports) > 0:
                row = [idx, f"{item['desc']}({self.tag[:12]})"]
            else:
                row = [idx, f"{item['desc']}"]
            report_str = self.gen_report_row(headers, item_type, key, precision, report_str, reports, row)
            data.append(row)
            report_str += "\n"
            for cr in compare_reports:
                row = [idx, f"{item['desc']}({cr['tag'][:12]})"]
                report_str = self.gen_report_row(headers, item_type, key, precision, report_str, cr['reports'], row, reports)
                data.append(row)
                report_str += "\n"
            idx += 1
        #
        headers.append("Total/Avg")
        print(f"Output (Tag: {self.tag}):")
        print(tabulate(data, headers=headers, tablefmt="pretty", stralign="left"))

    def gen_report_row(self, headers, item_type, key, precision, report_str, reports, row, compare_reports=None):
        total_value = 0
        for tn in self.test_names:
            item = reports[key][tn]
            compare_item = None
            if compare_reports:
                compare_item = compare_reports[key][tn]
            v = None
            if item['value'] is not None:
                if item_type == 'value':
                    v = round(item['value'], precision)
                    total_value += v
                    #
                    if compare_item:
                        cv = round(compare_item['value'], precision)
                        diff = round(v - cv, precision)
                        if cv < v:
                            v = f"{v} ({diff}) ↑"
                        elif cv > v:
                            v = f"{v} ({diff}) ↓"
                elif item_type == '%':  # %
                    v = round(item['value'] * 100, precision)
                    total_value += v
                    vs = f"{v} %"
                    if compare_item:
                        cv = round(compare_item['value'] * 100, precision)
                        diff = round(v - cv, precision)
                        if cv < v:
                            vs = f"{vs} ({diff}) ↑"
                        elif cv > v:
                            vs = f"{vs} ({diff}) ↓"
                    v = vs
            if v is None:
                v = item['value']
                total_value = v
            report_str += f"{v}  | "
            row.append(v)
        # total
        if total_value is not None:
            if item_type == 'value':
                total_value = f"{round(total_value, precision)} / {round(total_value / len(headers), precision)}"
            if item_type == '%':
                total_value = f"{round(total_value, precision)} % / {round(total_value / len(headers), precision)} %"
        row.append(total_value)
        return report_str

    #
    # def output_report(self, reports):
    #     report_str = f"\n-- Result (Tag: {self.tag}) --\n"
    #     idx = 1
    #     data = []
    #     headers = self.test_names
    #     for key in reports:
    #         item = reports[key][self.test_names[0]]
    #         report_str += f"{idx:02d}). {item['desc']}: "
    #         row = [idx, item['desc']]
    #         for tn in self.test_names:
    #             item = reports[key][tn]
    #             precision = item.get('precision', 2)
    #             item_type = item.get('type', 'value')
    #             if item_type == 'value':
    #                 v = round(item['value'], precision)
    #             elif item_type == '%':  # %
    #                 v = f"{round(item['value']*100, precision)} %"
    #             else:  # str
    #                 v = item['value']
    #             report_str += f"{v}  | "
    #             row.append(v)
    #         data.append(row)
    #
    #         report_str += "\n"
    #         idx += 1
    #     print(f"Output (Tag: {self.tag}):")
    #     print(tabulate(data, headers=headers, tablefmt="pretty"))

    def run_tester(self, test_config_path, test_name, script_path, log_path, print_log_type, result_value):
        pxt = PXTester(test_config_path=test_config_path, test_name=test_name, script_path=script_path,
                       log_path=log_path, print_log_type=print_log_type, test_result=result_value)
        pxt.execute("", sync=True)

    def get_script_path(self, args):
        if not args.scriptpath:
            return None
        if not os.path.isdir(args.scriptpath):
            return args.scriptpath
        if not args.tag:
            return None
        #
        dir_name = os.path.basename(args.scriptpath)
        for flag in ("_V", "_v", "@"):
            script_name = f"{dir_name}{flag}{args.tag}.py"
            script_path = os.path.join(args.scriptpath, script_name)
            if not os.path.isfile(script_path):
                script_path = None
            else:
                break

        return script_path



    def start_mp(self, args):
        # test_config_path = args.testconfig
        # test_name = args.testname, script_path = args.scriptpath,
        # log_path = args.logpath, print_log_type = args.printlogtype
        self.test_names = args.testname
        #
        results = {}
        manager = Manager()
        pool_args = []
        pool = Pool()
        script_path = self.get_script_path(args)
        if not script_path:
            print(f"ERROR: Invalid Script Path.")
            return 1

        for test_name in args.testname:
            tr = manager.Value(c_wchar_p, '')
            pool_args.append((args.testconfig, test_name, script_path, args.logpath, args.printlogtype, tr))
            results[test_name] = tr
        #
        for pa in pool_args:
            pool.apply_async(self.run_tester, pa)
        #
        pool.close()
        pool.join()
        reports = {}
        ri = {}
        for tn in results:
            res = json.loads(results[tn].value)
            ri[tn] = res
        #
        tag_data = dict(result=ri, utc_time=datetime.utcnow().isoformat())
        self.save_tag_data(self.tag, tag_data)
        self.__convert_report(reports, ri)
        compare_reports = []
        self.get_compare_reports(args, compare_reports)
        self.output_report(reports, compare_reports)

    def get_compare_reports(self, args, compare_reports):
        if args.compare:
            for c in args.compare:
                compare_data = self.load_tag_data(c)
                if compare_data:
                    r = {}
                    self.__convert_report(r, compare_data['result'])
                    compare_reports.append(dict(tag=c, reports=r))

    def __convert_report(self, reports, ri):
        for key in ri[self.test_names[0]]['report']:
            reports[key] = {}
            for tn in self.test_names:
                reports[key][tn] = ri[tn]['report'][key]

    def start_sp(self, args):
        self.test_names = args.testname
        #
        results = {}
        manager = Manager()
        pool_args = []
        script_path = self.get_script_path(args)
        if not script_path:
            print(f"ERROR: Invalid Script Path.")
            return 1

        for test_name in args.testname:
            tr = manager.Value(c_wchar_p, '')
            pool_args.append((args.testconfig, test_name, script_path, args.logpath, args.printlogtype, tr))
            results[test_name] = tr
        #
        for pa in pool_args:
            self.run_tester(*pa)
        #
        reports = {}
        ri = {}
        for tn in results:
            res = json.loads(results[tn].value)
            ri[tn] = res
        #
        tag_data = dict(result=ri, utc_time=datetime.utcnow().isoformat())
        self.save_tag_data(self.tag, tag_data)
        self.__convert_report(reports, ri)
        compare_reports = []
        self.get_compare_reports(args, compare_reports)
        self.output_report(reports, compare_reports)

    def compare_result(self, args):
        self.test_names = args.testname
        #
        #
        reports = {}
        tag_data = self.load_tag_data(self.tag)
        self.__convert_report(reports, tag_data['result'])
        #
        compare_reports = []
        self.get_compare_reports(args, compare_reports)
        self.output_report(reports, compare_reports)

def main(*args, **kwargs):
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--testconfig', type=str, required=True, help='Test config path')
    # parser.add_argument('-n', '--testname', type=str, required=True, help='Test name')
    parser.add_argument('-n', '--testname',  nargs='+', required=True, help='Test name')
    parser.add_argument('-s', '--scriptpath', type=str, required=False, help='Script path')
    parser.add_argument('-o', '--logpath', type=str, help='Log path')
    parser.add_argument('-p', '--printlogtype', nargs='+', default=['ea', 'report'], required=False,
                        help='Print log type,  account order ea report')
    parser.add_argument('-m', '--multiprocessing', type=MainApp.str2bool, default=True, help='Multiprocessing mode')
    parser.add_argument('-t', '--tag', type=str, help='Tag')
    parser.add_argument('-l', '--datafile', type=str, default='_pixiu_data.json', help='Data file name')
    parser.add_argument('-r', '--compare', nargs='+', help='Compare with the tags list')
    args = parser.parse_args()
    if args.scriptpath:
        if args.multiprocessing:
            MainApp(args).start_mp(args)
        else:
            MainApp(args).start_sp(args)
    elif len(args.compare) > 0 and args.tag:
        MainApp(args).compare_result(args)
    else:
        print("Error")

    # pxt = PXTester(test_config_path=args.testconfig, test_name=args.testname, script_path=args.scriptpath,
    #                log_path=args.logpath, print_log_type=args.printlogtype)
    # pxt.execute("", sync=True)


if __name__ == '__main__':
    main()




# ----------------------------------------------------------------------------------------------------------------------
#    DEFINES
# ----------------------------------------------------------------------------------------------------------------------
class TimeFrame(object):
    S1 = "s1"
    M1 = "m1"
    M5 = "m5"
    M15 = "m15"
    M30 = "m30"
    H1 = "h1"
    H4 = "h4"
    D1 = "d1"
    W1 = "w1"
    MN1 = "mn1"

class OrderCommand(object):
    ''''''
    BUY = 'BUY'
    SELL = 'SELL'
    BUYLIMIT = 'BUYLIMIT'
    SELLLIMIT = 'SELLLIMIT'
    BUYSTOP = 'BUYSTOP'
    SELLSTOP = 'SELLSTOP'
    #mt5
    BUYSTOPLIMIT = 'BUYSTOPLIMIT'
    SELLSTOPLIMIT = 'SELLSTOPLIMIT'
    CLOSEBY = 'CLOSEBY'


class OrderType(object):
    MARKET = 0
    LIMIT = 100
    STOP = 200

#
class RunModeValue(object):
    TEST = 'TEST'
    LIVE = 'LIVE'

class PositionType(object):
    SHORT = 'SHORT'
    LONG = 'LONG'

class OrderStatus(object):
    OPENED = 0
    CLOSED = 100
    PENDING = 200
    DELETED = 300
    CANCELLED = 400



from RestrictedPython.transformer import (RestrictingNodeTransformer, copy_locations,)
import ast



# https://rokujyouhitoma.hatenablog.com/entry/2018/06/05/063701
class EARestrictingNodeTransformer(RestrictingNodeTransformer):

    def inject_print_collector(self, node, position=0):
        #see: https://github.com/zopefoundation/RestrictedPython/blob/c1bc989e1fa060273594e39a86d3c4fb3ffe3a4b/src/RestrictedPython/transformer.py#L79
        print_used = self.print_info.print_used
        printed_used = self.print_info.printed_used

        if print_used or printed_used:
            # Add '_print = _print_(_getattr_)' add the top of a
            # function/module.
            _print = ast.Assign(
                targets=[ast.Name('_print', ast.Store())],
                value=ast.Call(
                    func=ast.Name("_print_", ast.Load()),
                    args=[ast.Name("_getattr_", ast.Load())],
                    keywords=[]))

            if isinstance(node, ast.Module):
                _print.lineno = position
                _print.col_offset = position
                ast.fix_missing_locations(_print)
            else:
                copy_locations(_print, node)

            node.body.insert(position, _print)

            # if not printed_used:
            #     self.warn(node, "Prints, but never reads 'printed' variable.")
            #
            # elif not print_used:
            #     self.warn(node, "Doesn't print, but reads 'printed' variable.")

    # def visit_Import(self, node):
    #     self.error(node, 'Import statements are not allowed.')
    # visit_ImportFrom = visit_Import

# source_code = "import this"
#
# try:
#     byte_code = compile_restricted(
#         source=source_code,
#         filename='<inline>',
#         mode='exec',
#         policy=EARestrictingNodeTransformer)
#     exec(byte_code)
# except SyntaxError as e:
#     raise e


#coding= utf-8



import os
import argparse
from datetime import (datetime, timedelta)
import sys
from unittest import (TestCase, TestLoader, TestSuite, TextTestRunner, skip, skipIf)
from pixiu.api.v1 import (TimeFrame, SymbolData)
from pixiu.tester import (EATester, )
import numpy as np
import time
import talib
import pandas as pd
import dateutil

debug_some_tests = False
# debug_some_tests = True

data = np.genfromtxt('usdchf_m1_20210315-0415.csv', delimiter=',',
                     dtype=[('s', object), ('t', float), ('o', float), ('h', float), ('c', float),
                            ('l', float), ('v', float), ('a', float), ('b', float), ], skip_header=1)
new_a = np.array(data,
                 dtype=[('s', object), ('t', float), ('o', float), ('h', float), ('c', float),
                        ('l', float), ('v', float), ('a', float), ('b', float), ])
#
# def to_tf_data(src, seconds):
#     temp_data = []
#     for sd in src:
#         s = sd['s']
#         t = int(sd['t'] / seconds) * seconds
#         o = src[src['t'] >= t]['o'][0]
#         h = src[(src['t'] >= t) & (src['t'] <= sd['t']) ]['h'].max()
#         l = src[(src['t'] >= t) & (src['t'] <= sd['t']) ]['l'].min()
#         c = sd['c']
#         a = sd['a']
#         b = sd['b']
#         v = src[(src['t'] >= t) & (src['t'] <= sd['t']) ]['v'].sum()
#         temp_data.append((s, t, o, h, c, l, v, a, b))
#     ret = np.array(temp_data,
#                  dtype=[('s', object), ('t', float), ('o', float), ('h', float), ('c', float),
#                         ('l', float), ('v', float), ('a', float), ('b', float), ])
#     return ret

def pandas_to_tf_data(src, timeframe):
    pd_tf_str = {TimeFrame.S1: '1s', TimeFrame.M1: '1T', TimeFrame.M5: '5T', TimeFrame.M15: '15T', TimeFrame.M30: '30T',
                 TimeFrame.H1: '1H', TimeFrame.H4: '4H', TimeFrame.D1: '1D', TimeFrame.S1: '1s', }
    tf_str = pd_tf_str.get(timeframe, None)
    if tf_str is None:
        return None
    da = pd.DataFrame(src)
    da['t'] = pd.to_datetime(da['t'], unit='s')
    da.set_index(pd.DatetimeIndex(da["t"]), inplace=True)
    logic = {'s': 'first',
             'o': 'first',
             'h': 'max',
             'c': 'last',
             'l': 'min',
             'v': 'sum',
             'a': 'last',
             'b': 'last'}
    new_da = da.resample(tf_str).apply(logic).dropna()
    new_da = new_da.reset_index()
    new_da['t'] = new_da['t'].values.astype(np.float64) // 10 ** 9
    ret = new_da.to_records(index=False)
    # ret = np.array(temp_data,
    #              dtype=[('s', object), ('t', float), ('o', float), ('h', float), ('c', float),
    #                     ('l', float), ('v', float), ('a', float), ('b', float), ])
    return ret



#
# df = pd.DataFrame()
# df = pd.read_csv('usdchf_m1_20210315-0415.csv', sep=',')
# pta.sma(df['c'], length=5)


class EATTester(EATester):
    def __init__(self, test_obj, params):
        super(EATTester, self).__init__(params)
        self.test_obj = test_obj


    def init_data(self):
        super(EATTester, self).init_data()
        self.safe_globals['t_check'] = self.t_check


    def on_load_ticks(self, *args, **kwargs):
        self.tick_info = self.get_data_info(symbol=self.symbol, timeframe=self.tick_timeframe,
                                            start_time=self.start_time,
                                            end_time=self.end_time)
        return 0

    def __update_execuate_log__(self, ticket, count=20, force=False):
        if force or time.time() - self.update_log_time > 2:  # 1s
            if count is not None:
                eidx = self.last_update_print_log_index + count
            else:
                eidx = None
            logs = self.print_logs[self.last_update_print_log_index:eidx]
            for l in logs:
                print(l)
            self.last_update_print_log_index += len(logs)


            #
            self.update_log_time = time.time()

    def on_end_tick(self, *args, **kwargs):
        self.__update_execuate_log__(self.ticket, count=20, force=False)
        return 0

    def on_end_execute(self, *args, **kwargs):
        self.__update_execuate_log__(self.ticket, None, force=True)
        return 0

    def GetSymbolData(self, symbol=None, timeframe=None):
        if symbol is None:
            symbol = self.symbol
        if timeframe is None:
            timeframe = self.tick_timeframe
        return self.get_data_info(symbol, timeframe)

    def __convert_db_price_data__(self, dbo_data, t_unit, t_count, desc):
        ''''''
        raw_data = []
        data = []

        for d in dbo_data:
            ask = d.ask if d.ask is not None else 0
            bid = d.bid if d.bid is not None else 0
            np = dict(t=d.time_frame, s=d.symbol, o=d.open_price,
                      h=d.high_price, c=d.close_price, l=d.low_price, v=d.volume,
                      a=ask, b=bid)
            raw_data.append(np)
            #
            np = (d.symbol, d.time_frame.timestamp(), d.open_price, d.high_price, d.close_price, d.low_price, d.volume,
                  ask, bid)
            data.append(np)

        return raw_data, data

    def symbol_datagetitem_index(self, data_index_ts, timeframe, timeframe_seconds, shift):
        t = int(self.tick_info[self.current_tick_index]['t']/ timeframe_seconds)*timeframe_seconds
        cidx = np.where(data_index_ts == t)[0][0]
        idx = cidx - shift
        return idx

    def symbol_datagetitem_callback(self, data, data_index_ts, timeframe, timeframe_seconds, shift, fail_value):
        idx = self.symbol_datagetitem_index(data_index_ts, timeframe, timeframe_seconds, shift)
        if idx < 0:
            return fail_value
        return data[idx]


    def get_data_info(self, symbol, timeframe=TimeFrame.M1,  start_time=None, end_time=None, last_count=None):
        """"""
        data = None
        if symbol is None:
            symbol = self.symbol
        symbol_tf = self.symbol_data.get(symbol, None)
        if symbol_tf:
            data = symbol_tf.get(timeframe, None)
        else:
            self.symbol_data[symbol] = {}
        if data is None:

            st = dateutil.parser.parse(self.start_time)
            et = dateutil.parser.parse(self.end_time)
            ary = new_a[(new_a['t'] >= st.timestamp()) & (new_a['t'] < et.timestamp())]
            #
            data = pandas_to_tf_data(ary, timeframe)
            self.symbol_data[symbol][timeframe] = data


        return self.symbol_data[symbol][timeframe]

    def t_check(self, name, value, *args, **kwargs):
        if name == "Close":
            self.test_obj.assertEqual(value, self.Close(), msg=f"{name}: {datetime.fromtimestamp(self.current_time())}: error!")

class PiXiuTests(TestCase):
    def setUp(self):
        self.test_result = ""
        self.spread_point = 15
        self.symbol = "USDCHF"
            #     values[datetime.fromtimestamp(new_a['t'][idx])][ti['name']] = ti['values'][idx]
        self.balance = 10000.0
        self.account = {'balance': self.balance,
                        'equity': self.balance,
                        'margin': 0,
                        'free_margin': self.balance,
                        'credit': 0.0,
                        'profit': 0.0,
                        'margin_level': 0,
                        #static
                        'leverage': 100,
                        'currency': 'USD',
                        'free_margin_mode': 0,
                        'stop_out_level': 0,
                        'stop_out_mode': 0,
                        'company': 'PIXIU',
                        'name': 'EATester',
                        'number': '000',
                        'server': 'EATester',
                        'trade_mode': 0,
                        'limit_orders': 500,
                        'margin_so_mode': 0,
                        'trade_allowed': True,
                        'trade_expert': 1,
                        'margin_so_call': 1.2,
                        'margin_so_so': 1.0,
                        'commission': 0.0,
                        }
        self.symbol_properties = dict(USDCHF={'symbol': 'USDCHF', 'spread': 2, 'digits': 5, 'stop_level': 0, 'volume_min': 0.01,
                                     'trade_contract_size': 100000, 'point': 0.00001,
                                     'currency_profit': 'CHF',
                                     'currency_base': 'USD',
                                     'currency_margin': 'USD', })
        self.eat_params = dict(symbol=self.symbol,
             start_time="2021-03-15",
             end_time="2021-04-16",
             tick_max_count=100,
             balance=self.balance,
             leverage=100,
             currency="USD",
             spread_point=self.spread_point,
             symbol_properties=self.symbol_properties,
             global_values=dict(
                 assertEqual=np.testing.assert_equal,
                 assertNotEqual=self.assertNotEqual,
                 assertIsNone=self.assertIsNone,
                 assertIsNotNone=self.assertIsNotNone,
                 assertTrue=self.assertTrue,
                 assertFalse=self.assertFalse,
                 exec_command=self.exec_command,
                 set_test_result=self.set_test_result,
             )
             )


    def tearDown(self):
        pass

    def exec_command(self):
        return 0

    def get_value_by_time(self, values, time, item_name):
        return values[time][item_name]

    def set_test_result(self, result):
        self.test_result = result

    def bbands_to_tuple_array(self, value):
        ret = []
        upper = value[0]
        middle = value[1]
        lower = value[2]
        for idx in range(upper.size):
            ret.append((upper[idx], middle[idx], lower[idx]))
        return ret

    def macd_to_tuple_array(self, value):
        ret = []
        macd = value[0]
        macdsignal = value[1]
        macdhist = value[2]
        for idx in range(macd.size):
            ret.append((macd[idx], macdsignal[idx], macdhist[idx]))
        return ret

    def stoch_to_tuple_array(self, value):
        ret = []
        slowk = value[0]
        slowd = value[1]
        for idx in range(slowk.size):
            ret.append((slowk[idx], slowd[idx]))
        return ret

    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_func_indicators(self):
        """Test EA Tester"""
        #
        timeperiod = 5
        matype = 0
        shift = 0
        test_items = [
            dict(name="ma",
                 values=talib.MA(new_a['c'], timeperiod=timeperiod, matype=matype),
                 statement=f"value=iMA(symbol_data.close, timeperiod={timeperiod}, matype={matype}, shift={shift})",
                 ),
            dict(name="ad",
                 values=talib.AD(new_a['h'], new_a['l'], new_a['c'], new_a['v']),
                 statement=f"value=iAD(symbol_data, shift={shift})",
                 ),
            dict(name="adx",
                 values=talib.ADX(new_a['h'], new_a['l'], new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iADX(symbol_data, timeperiod={timeperiod}, shift={shift})",
                 ),
            dict(name="atr",
                 values=talib.ATR(new_a['h'], new_a['l'], new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iATR(symbol_data, timeperiod={timeperiod}, shift={shift})",
                 ),
            dict(name="bands",
                 values=self.bbands_to_tuple_array(talib.BBANDS(new_a['c'], timeperiod=timeperiod, nbdevup=2, nbdevdn=2, matype=matype)),
                 statement=f"value=iBands(symbol_data.close, timeperiod={timeperiod}, nbdevup=2, nbdevdn=2, matype={matype}, shift={shift})",
                 ),
            dict(name="cci",
                 values=talib.CCI(new_a['h'], new_a['l'], new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iCCI(symbol_data, timeperiod={timeperiod}, shift={shift})",
                ),
            dict(name="chaikin",
                 values=talib.ADOSC(new_a['h'], new_a['l'], new_a['c'], new_a['v'], fastperiod=5, slowperiod=10),
                 statement=f"value=iChaikin(symbol_data, fastperiod=5, slowperiod=10, shift={shift})",
                ),
            dict(name="dema",
                 values=talib.DEMA(new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iDEMA(symbol_data.close, timeperiod={timeperiod}, shift={shift})",
                ),
            dict(name="momentum",
                 values=talib.MOM(new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iMomentum(symbol_data.close, timeperiod={timeperiod}, shift={shift})",
                ),
            dict(name="mfi",
                 values=talib.MFI(new_a['h'], new_a['l'], new_a['c'], new_a['v'], timeperiod=timeperiod),
                 statement=f"value=iMFI(symbol_data, timeperiod={timeperiod}, shift={shift})",
                ),
            dict(name="macd",
                 values=self.macd_to_tuple_array(talib.MACD(new_a['c'], fastperiod=5, slowperiod=10, signalperiod=7)),
                 statement=f"value=iMACD(symbol_data.close, fastperiod=5, slowperiod=10, signalperiod=7, shift={shift})",
                 ),
            dict(name="obv",
                 values=talib.OBV(new_a['c'], new_a['v']),
                 statement=f"value=iOBV(symbol_data.close, shift={shift})",
                 ),
            dict(name="sar",
                 values=talib.SAR(new_a['h'], new_a['l'], acceleration=0.02, maximum=0.2),
                 statement=f"value=iSAR(symbol_data, acceleration=0.02, maximum=0.2, shift={shift})",
                 ),
            dict(name="rsi",
                 values=talib.RSI(new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iRSI(symbol_data.close, timeperiod={timeperiod}, shift={shift})",
                 ),
            dict(name="stddev",
                 values=talib.STDDEV(new_a['c'], timeperiod=timeperiod, nbdev=2),
                 statement=f"value=iStdDev(symbol_data.close, timeperiod={timeperiod}, nbdev=2, shift={shift})",
                 ),
            dict(name="stochastic",
                 values=self.stoch_to_tuple_array(talib.STOCH(new_a['h'], new_a['l'], new_a['c'], fastk_period=5, slowk_period=3,
                                    slowk_matype=0, slowd_period=3, slowd_matype=0)),
                 statement=f"value=iStochastic(symbol_data, fastk_period=5, slowk_period=3, slowk_matype=0, slowd_period=3, slowd_matype=0, shift={shift})",
                 ),
            dict(name="tema",
                 values=talib.TEMA(new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iTEMA(symbol_data.close, timeperiod={timeperiod}, shift={shift})",
                 ),
            dict(name="wpr",
                 values=talib.WILLR(new_a['h'], new_a['l'], new_a['c'], timeperiod=timeperiod),
                 statement=f"value=iWPR(symbol_data, timeperiod={timeperiod}, shift={shift})",
                 ),
        ]
        values = {}
        # ma_timeperiod = 5
        # ma_matype = 1
        # ma_values = talib.MA(new_a['c'], timeperiod=ma_timeperiod, matype=ma_matype)
        for idx in range(new_a.size):
            values[datetime.fromtimestamp(new_a['t'][idx])] = {}
            for ti in test_items:
                values[datetime.fromtimestamp(new_a['t'][idx])][ti['name']] = ti['values'][idx]

        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_indicators.py")
        self.eat_params['global_values'].update(dict(valid_symbol=self.symbol,
                                                     valid_values=values, valid_account=self.account,
                                                     valid_symbols=self.symbol_properties,
                                                     get_value_by_time=self.get_value_by_time))
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")


    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_func_load_lib(self):
        """Test EA Tester"""
        #
        values = {}
        self.eat_params['script_libs'] = []
        for i in range(3):
            lib_path = os.path.abspath(f"scripts/v1/ts_lib{i}.py")
            self.eat_params['script_libs'].append(dict(name=f'test_lib{i}', version='1.0', path=lib_path,
                                                code=open(lib_path).read(), metadata={}))
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_lib_load.py")
        self.eat_params['global_values'].update(dict(valid_shift=0, valid_symbol=self.symbol,
                                                     valid_values=values, valid_account=self.account,
                                                     valid_symbols=self.symbol_properties,
                                                     get_value_by_time=self.get_value_by_time))
        self.eat_params['log_path'] = 'tmp_log.txt'
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")

    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_func_base(self):
        """Test EA Tester"""
        # ma_timeperiod = 5
        # ma_matype = 1
        # ma_values = talib.MA(new_a['c'], timeperiod=ma_timeperiod, matype=ma_matype)
        values = {}
        for idx in range(new_a.size):
            a = new_a[idx]
            close = a['c']
            ask = a['a']
            bid = a['b']
            if bid == 0:
                bid = close
            if ask == 0:
                ask = bid + (self.spread_point * self.symbol_properties[self.symbol]['point'])
            values[datetime.fromtimestamp(new_a['t'][idx])] = dict(open=a['o'], close=close,
                                                                   high=a['h'], low=a['l'],
                                                                   ask=ask, bid=bid,
                                                                   volume=a['v'],
                                                                   time=datetime.fromtimestamp(a['t'])
                                                                   )

        #
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_base.py")
        self.eat_params['global_values'].update(dict(valid_shift=0, valid_symbol=self.symbol,
                                                     valid_values=values, valid_account=self.account,
                                                     valid_symbols=self.symbol_properties,
                                                     get_value_by_time=self.get_value_by_time))
        self.eat_params['log_path'] = 'tmp_log.txt'
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")


    def get_timeframe_value_by_time(self, timeframe, v_time, item_name):
        tfd = self.timeframe_data.get(timeframe, None)
        seconds = tfd['seconds']
        ts = int(v_time.timestamp() / seconds) * seconds
        a = tfd['data'][tfd['data']['t'] == ts][0]
        close = a['c']
        ask = a['a']
        bid = a['b']
        # if bid == 0:
        #     bid = close
        # if ask == 0:
        #     ask = bid + (self.spread_point * self.symbol_properties[self.symbol]['point'])
        v = dict(open=a['o'], close=close,
                            high=a['h'], low=a['l'],
                                                               ask=ask, bid=bid,
                                                               volume=a['v'],
                                                               time=datetime.fromtimestamp(a['t'])
                                                               )
        return v[item_name]

    def init_values(self):
        self.values = {}
        self.timeframe_data = {TimeFrame.S1: {'seconds': 1}, TimeFrame.M1: {'seconds': 60}, TimeFrame.M5: {'seconds': 300},
                          TimeFrame.M15: {'seconds': 900}, TimeFrame.M30: {'seconds': 1800},
                          TimeFrame.H1: {'seconds': 3600}, TimeFrame.H4: {'seconds': 14400},
                          TimeFrame.D1: {'seconds': 86400}}
        for tf in self.timeframe_data:
            self.timeframe_data[tf]['data'] = pandas_to_tf_data(new_a, tf)
            self.values[tf] = {}

        for idx in range(new_a.size):
            a = new_a[idx]
            close = a['c']
            ask = a['a']
            bid = a['b']
            # if bid == 0:
            #     bid = close
            # if ask == 0:
            #     ask = bid + (self.spread_point * self.symbol_properties[self.symbol]['point'])
            self.values[datetime.fromtimestamp(new_a['t'][idx])] = dict(open=a['o'], close=close,
                                                                   high=a['h'], low=a['l'],
                                                                   ask=ask, bid=bid,
                                                                   volume=a['v'],
                                                                   time=datetime.fromtimestamp(a['t'])
                                                                   )


    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_func_symbol_data(self):
        """Test EA Tester"""
        # ma_timeperiod = 5
        # ma_matype = 1
        # ma_values = talib.MA(new_a['c'], timeperiod=ma_timeperiod, matype=ma_matype)
        self.init_values()

        #tick > 1day
        self.eat_params['tick_max_count'] = 1500
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_symbol_data.py")
        self.eat_params['global_values'].update(dict(valid_shift=0, valid_symbol=self.symbol,
                                                     valid_values=self.values, valid_account=self.account,
                                                     valid_symbols=self.symbol_properties,
                                                     get_value_by_time=self.get_value_by_time,
                                                     get_timeframe_value_by_time=self.get_timeframe_value_by_time),
                                                )
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")



    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_order_market(self):
        """Test EA Tester"""
        #
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_order_market.py")
        # self.eat_params['global_values']['valid_buylimit_price'] = buylimit_price
        # self.eat_params['global_values']['valid_buylimit_time'] = buylimit_time
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")


    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_order_buylimit(self):
        """Test EA Tester"""
        #
        a = new_a['a'][0]
        for idx in range(new_a.size):
            p = new_a[idx]['a']
            if p < a:
                buylimit_price = p
                buylimit_time = datetime.fromtimestamp(new_a[idx]['t'])
                break
        #
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_order_buylimit.py")
        self.eat_params['global_values']['valid_buylimit_price'] = buylimit_price
        self.eat_params['global_values']['valid_buylimit_time'] = buylimit_time
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")


    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_order_buystop(self):
        """Test EA Tester"""
        #
        #
        a = new_a['a'][0]
        for idx in range(new_a.size):
            p = new_a[idx]['a']
            if p > a:
                buystop_price = p
                buystop_time = datetime.fromtimestamp(new_a[idx]['t'])
                break
        #
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_order_buystop.py")
        self.eat_params['global_values']['valid_buystop_price'] = buystop_price
        self.eat_params['global_values']['valid_buystop_time'] = buystop_time
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")


    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_order_selllimit(self):
        """Test EA Tester"""
        #
        b = new_a['b'][0]
        for idx in range(new_a.size):
            p = new_a[idx]['b']
            if p > b:
                selllimit_price = p
                selllimit_time = datetime.fromtimestamp(new_a[idx]['t'])
                break
        #
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_order_selllimit.py")
        self.eat_params['global_values']['valid_selllimit_price'] = selllimit_price
        self.eat_params['global_values']['valid_selllimit_time'] = selllimit_time
        eatt = EATTester(self, self.eat_params)
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")


    @skipIf(debug_some_tests, "debug some tests")
    def test_ea_tester_order_sellstop(self):
        """Test EA Tester"""
        b = new_a['b'][0]
        for idx in range(new_a.size):
            p = new_a[idx]['b']
            if p < b:
                sellstop_price = p
                sellstop_time = datetime.fromtimestamp(new_a[idx]['t'])
                break
        #
        self.eat_params['script_path'] = os.path.abspath("scripts/v1/ts_order_sellstop.py")
        self.eat_params['global_values']['valid_sellstop_price'] = sellstop_price
        self.eat_params['global_values']['valid_sellstop_time'] = sellstop_time

        eatt = EATTester(self, self.eat_params)
        # with self.assertRaisesRegex(Exception, "EID_EAT_TEST_STOP"):
        eatt.execute("123456", sync=True)
        self.assertEqual(self.test_result, "OK")




# --------------------------------------------------------------------
def test(test_cases=[]):
    pixiuTests = TestLoader().loadTestsFromTestCase(PiXiuTests)

    #
    test_suites = TestSuite([pixiuTests])
    tmp_ts = TestSuite()
    if len(test_cases) > 0:
        for ts in test_suites:
            for t in ts:
                if t._testMethodName in test_cases:
                    tmp_ts.addTest(t)
        test_suites = tmp_ts
    #
    test_result = TextTestRunner(verbosity=2).run(test_suites)

    return test_result.wasSuccessful()

def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--test_cases', nargs='+', default=[], required=False,
                        help='test cases, ex: case1 case2 case3 ... caseN')
    args = parser.parse_args()
    ret = test(test_cases=args.test_cases)
    if ret:
        sys.exit(0)
    else:
        sys.exit(1)





# SetMetadata("copyright", "2022")

def test_timezone():
    tz = timezone('EST')
    d = Time().astimezone(tz)
    assertIsNotNone(d)


assertTrue(RunMode() in (RunModeValue.TEST, RunModeValue.LIVE))

print("ts_base: ", "test print", "1", "2", "3", "4", "5")
assertEqual(TimeFrame.S1, "s1")
assertEqual(TimeFrame.M1, "m1")
assertEqual(TimeFrame.M5, "m5")
assertEqual(TimeFrame.M15, "m15")
assertEqual(TimeFrame.M30, "m30")
assertEqual(TimeFrame.H1, "h1")
assertEqual(TimeFrame.H4, "h4")
assertEqual(TimeFrame.D1, "d1")
assertEqual(TimeFrame.W1, "w1")
assertEqual(TimeFrame.MN1, "mn1")

#OrderType
assertEqual(OrderType.MARKET, 0)
assertEqual(OrderType.LIMIT, 100)
assertEqual(OrderType.STOP, 200)

#OrderCommand
assertEqual(OrderCommand.BUY, 'BUY')
assertEqual(OrderCommand.SELL, 'SELL')
assertEqual(OrderCommand.BUYLIMIT, 'BUYLIMIT')
assertEqual(OrderCommand.SELLLIMIT, 'SELLLIMIT')
assertEqual(OrderCommand.BUYSTOP, 'BUYSTOP')
assertEqual(OrderCommand.SELLSTOP, 'SELLSTOP')
assertEqual(OrderCommand.BUYSTOPLIMIT, 'BUYSTOPLIMIT')
assertEqual(OrderCommand.SELLSTOPLIMIT, 'SELLSTOPLIMIT')
assertEqual(OrderCommand.CLOSEBY, 'CLOSEBY')

#symbol
assertEqual(Symbol(), valid_symbol)

c_time = Time(valid_shift)

#Test basic values (Cache)
assertEqual(c_time, get_value_by_time(valid_values, c_time, "time"))
assertEqual(Open(valid_shift), get_value_by_time(valid_values, c_time, "open"))
assertEqual(Close(valid_shift), get_value_by_time(valid_values, c_time, "close"))
assertEqual(High(valid_shift), get_value_by_time(valid_values, c_time, "high"))
assertEqual(Low(valid_shift), get_value_by_time(valid_values, c_time, "low"))
assertEqual(Ask(valid_shift), get_value_by_time(valid_values, c_time, "ask"), f'@ {c_time}')
assertEqual(Bid(valid_shift), get_value_by_time(valid_values, c_time, "bid"))
assertEqual(Volume(valid_shift), get_value_by_time(valid_values, c_time, "volume"))

#Account
account = GetAccount()
assertEqual(account.balance, valid_account['balance'])
assertEqual(account.equity, valid_account['equity'])
assertEqual(account.margin, valid_account['margin'])
assertEqual(account.free_margin, valid_account['free_margin'])
assertEqual(account.credit, valid_account['credit'])
assertEqual(account.profit, valid_account['profit'])
assertEqual(account.margin_level, valid_account['margin_level'])
assertEqual(account.leverage, valid_account['leverage'])
assertEqual(account.currency, valid_account['currency'])
assertEqual(account.free_margin_mode, valid_account['free_margin_mode'])
assertEqual(account.stop_out_level, valid_account['stop_out_level'])
assertEqual(account.stop_out_mode, valid_account['stop_out_mode'])
assertEqual(account.company, valid_account['company'])
assertEqual(account.name, valid_account['name'])
assertEqual(account.number, valid_account['number'])
assertEqual(account.server, valid_account['server'])
assertEqual(account.trade_mode, valid_account['trade_mode'])
assertEqual(account.limit_orders, valid_account['limit_orders'])
assertEqual(account.margin_so_mode, valid_account['margin_so_mode'])
assertEqual(account.trade_allowed, valid_account['trade_allowed'])
assertEqual(account.trade_expert, valid_account['trade_expert'])
assertEqual(account.margin_so_call, valid_account['margin_so_call'])
assertEqual(account.margin_so_so, valid_account['margin_so_so'])
assertEqual(account.commission, valid_account['commission'])

#Symbol
symbol = GetSymbol(Symbol())
assertEqual(symbol.name, valid_symbols[Symbol()]['symbol'])
assertEqual(symbol.spread, valid_symbols[Symbol()]['spread'])
assertEqual(symbol.digits, valid_symbols[Symbol()]['digits'])
assertEqual(symbol.stop_level, valid_symbols[Symbol()]['stop_level'])
assertEqual(symbol.volume_min, valid_symbols[Symbol()]['volume_min'])
assertEqual(symbol.trade_contract_size, valid_symbols[Symbol()]['trade_contract_size'])
assertEqual(symbol.point, valid_symbols[Symbol()]['point'])
assertEqual(symbol.currency_profit, valid_symbols[Symbol()]['currency_profit'])
assertEqual(symbol.currency_base, valid_symbols[Symbol()]['currency_base'])
assertEqual(symbol.currency_margin, valid_symbols[Symbol()]['currency_margin'])


#Data
for scope in (DataScope.EA, DataScope.EA_VERSION, DataScope.ACCOUNT, DataScope.EA_SETTIGNS):
    name = hashlib.md5(str(uuid.uuid4()).encode("utf-8")).hexdigest()
    for i in range(3):
        data = dict(value=UID())
        SaveData(scope=scope, name=name, data=data)
        d = LoadData(scope=scope, name=name)
        assertEqual(d['value'], data['value'], f"scope={scope}, i={i}")
    DeleteData(scope=scope, name=name)
    d = LoadData(scope=scope, name=name)
    assertIsNone(d, f"scope={scope}")

# Test functions
test_timezone()

#
set_test_result("OK")
#
StopTester()


EID_OK = 0
EID_EAT_ERROR = 2104301300
EID_EAT_INVALID_ORDER_VOLUME = 2104301301
EID_EAT_INVALID_MARKET_ORDER_OPEN_PRICE = 2104301302
EID_EAT_INVALID_LIMIT_ORDER_OPEN_PRICE = 2104301303
EID_EAT_INVALID_STOP_ORDER_OPEN_PRICE = 2104301304
EID_EAT_INVALID_ORDER_TYPE = 2104301305
EID_EAT_INVALID_ORDER_CLOSE_PRICE = 2104301306
EID_EAT_INVALID_ORDER_STOP_LOSS = 2104301307
EID_EAT_INVALID_ORDER_TAKE_PROFIT = 2104301308
EID_EAT_NOT_ENOUGH_MONEY = 2104301309
EID_EAT_INVALID_ORDER_TICKET = 2104301310
EID_EAT_TEST_STOP = 2104301311
EID_EAT_EA_DEAD = 2104301312
EID_EAT_LIMIT_ORDERS = 2104301313
EID_EAT_INVALID_SYMBOL = 2104301314
EID_EAT_INVALID_DATA_FORMAT = 21043015
EID_EAT_INVALID_DATA_LENGTH = 21043016


class PXErrorCode(Exception):
    ''''''
    def __init__(self, errid):
        self.errid = errid




POINT = SymbolInfo("point")
DIGITS = int(SymbolInfo("digits"))

ood = GetOpenedOrderUIDs()
if len(ood) == 0:
    pod = GetPendingOrderUIDs()
    assertIsNotNone(pod)
    if len(pod) == 0:
        #
        volume = 0.01

        # Buy all params
        stop_loss = round(Ask() + 15 * POINT, DIGITS)
        take_profit = round(Bid() - 30 * POINT, DIGITS)
        comment = "test sellstop"
        magic_number = 4301250
        # failed
        price = Bid()
        errid, result = Sell(volume=0.01, type=OrderType.STOP, price=price, stop_loss=stop_loss,
                            take_profit=take_profit, magic_number=magic_number,
                            symbol=Symbol(), slippage=3, arrow_color="white")
        assertEqual(errid, EID_EAT_INVALID_STOP_ORDER_OPEN_PRICE)

        # failed
        price = round(Bid() + 10 * POINT, DIGITS)
        errid, result = Sell(volume=0.01, type=OrderType.STOP, price=price, stop_loss=stop_loss,
                            take_profit=take_profit, magic_number=magic_number,
                            symbol=Symbol(), slippage=3, arrow_color="white")
        assertEqual(errid, EID_EAT_INVALID_STOP_ORDER_OPEN_PRICE)

        #test open, modify, close
        price = valid_sellstop_price
        errid, result = Sell(volume=0.01, type=OrderType.STOP, price=price, stop_loss=stop_loss,
                            take_profit=take_profit, magic_number=magic_number,
                            symbol=Symbol(), slippage=3, arrow_color="white")
        assertEqual(errid, 0)
        assertIsNotNone(result)
        errid = exec_command()
        assertEqual(errid, 0)
        oo = GetPendingOrderUIDs()
        assertTrue(result['order_uid'] in oo)
        order = GetOrder(result['order_uid'])
        assertEqual(order.status, OrderStatus.PENDING)

        # modify
        stop_loss = round(stop_loss + 15 * POINT, DIGITS)
        take_profit = round(take_profit - 30 * POINT, DIGITS)
        errid, result = ModifyOrder(result['order_uid'], price=Bid()+1*POINT, stop_loss=stop_loss, take_profit=take_profit)
        assertEqual(errid, EID_EAT_INVALID_STOP_ORDER_OPEN_PRICE)
        errid, result = ModifyOrder(result['order_uid'], price=Bid()-1*POINT, stop_loss=stop_loss, take_profit=take_profit)
        assertEqual(errid, 0)
        assertIsNotNone(result)
        errid = exec_command()
        assertEqual(errid, 0)

        #
        order = GetOrder(result['order_uid'])
        assertEqual(order.symbol, Symbol())
        assertEqual(order.uid, result['order_uid'])
        assertEqual(order.volume, volume)
        assertEqual(order.stop_loss, stop_loss)
        assertEqual(order.take_profit, take_profit)
        assertIsNone(order.close_time)
        assertIsNone(order.close_price)
        assertEqual(order.status, OrderStatus.PENDING)

        #close
        errid, close_result = CloseOrder(result['order_uid'], volume=volume, price=Ask())
        assertEqual(errid, 0)
        assertEqual(close_result['order_uid'], result['order_uid'])
        errid = exec_command()
        assertEqual(errid, 0)
        order = GetOrder(result['order_uid'])
        assertEqual(order.status, OrderStatus.CANCELLED)

        # The order was closed ?
        oo = GetPendingOrderUIDs()
        assertFalse(result['order_uid'] in oo)

        #
        price = valid_sellstop_price
        errid, result = Sell(volume=0.01, type=OrderType.STOP, price=price, stop_loss=stop_loss,
                            take_profit=take_profit, magic_number=magic_number,
                            symbol=Symbol(), slippage=3, arrow_color="white")
        assertEqual(errid, 0)
        assertIsNotNone(result)
        errid = exec_command()
        assertEqual(errid, 0)
        order = GetOrder(result['order_uid'])
        assertEqual(order.status, OrderStatus.PENDING)

else:
    pass
ctime = Time()
if ctime == valid_sellstop_time:
    pod = GetPendingOrderUIDs()
    assertEqual(len(pod), 0)
    assertEqual(len(ood), 1)
    # order = GetOrder(order_uid)
    # assertEqual(order.symbol, Symbol())
    # assertEqual(order.order_uid, order_uid)
    # assertEqual(order.open_time, ctime)
    # assertEqual(order.volume, volume)
    # assertEqual(order.stop_loss, stop_loss)
    # assertEqual(order.take_profit, take_profit)
    # assertEqual(order.comment, "open")
    # assertEqual(order.magic_number, magic_number)
    # assertIsNone(order.close_time)
    # assertEqual(order.close_price, nan)
    set_test_result("OK")
    StopTester()


